from typing import Union, Dict, Any
import abc

from fabric import Connection
from invoke import Context

LOCAL_ADDRS = [
    'local',
    'localhost',
]


class Host(abc.ABC):
    addr: str
    context: Dict[str, Any]

    def __init__(
        self,
        addr: str,
        ssh_user: str = None, ssh_password: str = None,
        ssh_connect_timeout: int = 10,
        **context
     ):
        """
        Defined host to operate in
        :param addr: user@host, host, ip for remote, one if LOCAL_ADDRS for local execution
        :param context: Some context vars for use in runtime
        """
        self.addr = addr
        self.context = context
        self.ssh_user = ssh_user
        self.ssh_password = ssh_password
        self.ssh_connect_timeout = ssh_connect_timeout

    def is_connection_local(self):
        return self.addr in LOCAL_ADDRS

    def connect(self) -> Union[Connection, Context]:
        if self.is_connection_local():
            # Host is local machine
            return Context()
        else:
            # Host is remote ssh machine
            return Connection(
                self.addr,
                user=self.ssh_user,
                connect_timeout=self.ssh_connect_timeout,
                connect_kwargs={
                    'password': self.ssh_password,
                }
            )

    @property
    def host(self) -> str:
        # Remove user and port parts

        h = self.addr
        if '@' in self.addr:
            h = h.split("@", maxsplit=1)[1]

        if ':' in self.addr:
            h = h.split(":", maxsplit=1)[0]

        return h

    def __str__(self):
        return f"🖥 {self.addr}"

    def __hash__(self):
        return hash(self.addr)



from typing import Optional

import invoke

from carnival import cmd
from carnival.utils import log


def get_installed_version(pkgname: str) -> Optional[str]:
    """
    Get installed package version
    Returns None if package not installed
    """
    try:
        ver = cmd.cli.run(f"dpkg -s {pkgname} | grep Version", hide='both').stdout
    except invoke.exceptions.UnexpectedExit as ex:
        stderr = ex.result.tail("stderr")
        if 'is not installed and no information is available' in stderr:
            return None
        raise ValueError from ex

    _, v = ver.split("Version: ", maxsplit=1)
    return v.strip()


def is_pkg_installed(pkgname: str, version=None) -> bool:
    """
    Check is package installed?
    If version not specified - check any version
    """

    pkgver = get_installed_version(pkgname)
    if version is None and pkgver is not None:
        return True

    if version is not None and pkgver == version:
        return True

    return False


def force_install(pkgname, version=None, update=False):
    """
    Install apt package
    """
    if version:
        pkgname = f"{pkgname}={version}"

    if update:
        cmd.cli.run("sudo apt-get update", pty=True)

    cmd.cli.run(f"sudo apt-get install -y {pkgname}", pty=True)


def install(pkgname, version=None, update=True) -> bool:
    """
    Install apt package if not installed
    Returns true if installed, false if was already installed
    """
    if is_pkg_installed(pkgname, version):
        if version:
            log(f"{pkgname}={version} already installed")
        else:
            log(f"{pkgname} already installed")
        return False
    force_install(pkgname=pkgname, version=version, update=update)
    return True


def install_multiple(*pkg_names: str, update=True):
    if all([is_pkg_installed(x) for x in pkg_names]):
        log(f"{','.join(pkg_names)} already installed")
        return False

    if update:
        cmd.cli.run("sudo apt-get update", pty=True)

    for pkg in pkg_names:
        install(pkg, update=False)
    return True



from typing import List

from carnival import cmd


def set_password(username: str, password: str):
    raise NotImplementedError  # TODO


def ssh_append_id(ssh_key: str, keys_file=".ssh/authorized_keys"):
    ssh_key = ssh_key.strip()

    cmd.cli.run("mkdir -p ~/.ssh")
    cmd.cli.run("chmod 700 ~/.ssh")
    cmd.cli.run(f"touch {keys_file}")

    if not cmd.transfer.is_file_contains(keys_file, ssh_key, escape=True):
        cmd.cli.run(f"echo '{ssh_key}' >> {keys_file}")


def ssh_ensure_keys(ssh_keys: List[str]):
    for ssh_key in ssh_keys:
        ssh_append_id(ssh_key)



from typing import List, Union

import re

from carnival import global_context, Step

from carnival.host import Host


def _underscore(word: str) -> str:
    # https://github.com/jpvanhal/inflection/blob/master/inflection.py
    word = re.sub(r"([A-Z]+)([A-Z][a-z])", r'\1_\2', word)
    word = re.sub(r"([a-z\d])([A-Z])", r'\1_\2', word)
    word = word.replace("-", "_")
    return word.lower()


class Task:
    name: str = ""

    @classmethod
    def get_name(cls) -> str:
        return cls.name if cls.name else _underscore(cls.__name__)

    def __init__(self, dry_run: bool):
        self.dry_run = dry_run

    def step(self, roles: Union[Step, List[Step]], hosts: Union[Host, List[Host]]):
        if not isinstance(roles, list) and not isinstance(roles, tuple):
            roles = [roles, ]

        if not isinstance(hosts, list) and not isinstance(hosts, tuple):
            hosts = [hosts, ]

        for host in hosts:
            global_context.set_context(host)

            for role in roles:
                role_name = _underscore(role.__class__.__name__)
                print(f"💃💃💃 Running {role_name} at {host}")
                if not self.dry_run:
                    role.run_with_context(host=host)

    def run(self, **kwargs):
        raise NotImplementedError



from typing import Union

from fabric import Connection
from invoke import Context

from carnival.host import Host


# noinspection PyTypeChecker
conn: Union[Connection, Context] = None
# noinspection PyTypeChecker
host: Host = None


def set_context(h: Host):
    global conn
    global host
    conn = h.connect()
    host = h



import os
import sys
from typing import Iterable, Type, Dict

import click

from carnival.task import Task


def load_tasks_file(tasks_file: str) -> Dict[str, Type[Task]]:
    try:
        __import__(os.path.splitext(tasks_file)[0])
    except ModuleNotFoundError:
        return {}

    tasks: Dict[str, Type[Task]] = {}

    for task_class in Task.__subclasses__():
        tasks[task_class.get_name()] = task_class

    return tasks


def main():
    sys.path.insert(0, os.getcwd())
    carnival_file = os.getenv("CARNIVAL_FILE", "carnival_file.py")
    try:
        task_types = load_tasks_file(carnival_file)
    except FileNotFoundError:
        print(f"Carnival file ({carnival_file}) not found.")
        return

    @click.command()
    @click.option('-d', '--dry_run', is_flag=True, default=False, help="Simulate run")
    @click.argument('tasks', required=True, type=click.Choice(task_types.keys()), nargs=-1)
    def cli(dry_run: bool, tasks: Iterable[str]):
        for task in tasks:
            task_types[task](dry_run=dry_run).run()
    cli()
